@@ -33,15 +33,16 @@ import codecs
 import datetime
 import functools
 import inspect
+import io
 import itertools
 import json
 import uuid
+import warnings
+from xmlrpc import client as xmlrpclib
 
 from oslo_utils import encodeutils
 from oslo_utils import importutils
 from oslo_utils import timeutils
-import six
-import six.moves.xmlrpc_client as xmlrpclib
 
 ipaddress = importutils.try_import("ipaddress")
 netaddr = importutils.try_import("netaddr")
@@ -52,12 +53,12 @@ _nasty_type_tests = [inspect.ismodule, inspect.isclass, inspect.ismethod,
                      inspect.iscode, inspect.isbuiltin, inspect.isroutine,
                      inspect.isabstract]
 
-_simple_types = ((six.text_type,) + six.integer_types
-                 + (type(None), bool, float))
+_simple_types = (str, int, type(None), bool, float)
 
 
 def to_primitive(value, convert_instances=False, convert_datetime=True,
-                 level=0, max_depth=3, encoding='utf-8'):
+                 level=0, max_depth=3, encoding='utf-8',
+                 fallback=None):
     """Convert a complex object into primitives.
 
     Handy for JSON serialization. We can optionally handle instances,
@@ -70,12 +71,22 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
 
     Therefore, ``convert_instances=True`` is lossy ... be aware.
 
+    If the object cannot be converted to primitive, it is returned unchanged
+    if fallback is not set, return fallback(value) otherwise.
+
+    .. versionchanged:: 2.22
+       Added *fallback* parameter.
+
     .. versionchanged:: 1.3
        Support UUID encoding.
 
     .. versionchanged:: 1.6
        Dictionary keys are now also encoded.
     """
+    orig_fallback = fallback
+    if fallback is None:
+        fallback = str
+
     # handle obvious types first - order of basic types determined by running
     # full tests on nova project, resulting in the following counts:
     # 572754 <type 'NoneType'>
@@ -92,10 +103,8 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
     if isinstance(value, _simple_types):
         return value
 
-    if isinstance(value, six.binary_type):
-        if six.PY3:
-            value = value.decode(encoding=encoding)
-        return value
+    if isinstance(value, bytes):
+        return value.decode(encoding=encoding)
 
     # It's not clear why xmlrpclib created their own DateTime type, but
     # for our purposes, make it a datetime type which is explicitly
@@ -110,31 +119,27 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
             return value
 
     if isinstance(value, uuid.UUID):
-        return six.text_type(value)
+        return str(value)
 
     if netaddr and isinstance(value, (netaddr.IPAddress, netaddr.IPNetwork)):
-        return six.text_type(value)
+        return str(value)
 
     if ipaddress and isinstance(value,
                                 (ipaddress.IPv4Address,
                                  ipaddress.IPv6Address)):
-        return six.text_type(value)
+        return str(value)
+
+    # For exceptions, return the 'repr' of the exception object
+    if isinstance(value, Exception):
+        return repr(value)
 
     # value of itertools.count doesn't get caught by nasty_type_tests
     # and results in infinite loop when list(value) is called.
     if type(value) == itertools.count:
-        return six.text_type(value)
+        return fallback(value)
 
     if any(test(value) for test in _nasty_type_tests):
-        return six.text_type(value)
-
-    # FIXME(vish): Workaround for LP bug 852095. Without this workaround,
-    #              tests that raise an exception in a mocked method that
-    #              has a @wrap_exception with a notifier will fail. If
-    #              we up the dependency to 0.5.4 (when it is released) we
-    #              can remove this workaround.
-    if getattr(value, '__module__', None) == 'mox':
-        return 'mock'
+        return fallback(value)
 
     if level > max_depth:
         return None
@@ -147,7 +152,8 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
                                       convert_datetime=convert_datetime,
                                       level=level,
                                       max_depth=max_depth,
-                                      encoding=encoding)
+                                      encoding=encoding,
+                                      fallback=orig_fallback)
         if isinstance(value, dict):
             return {recursive(k): recursive(v)
                     for k, v in value.items()}
@@ -156,7 +162,7 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
         # Python 3 does not have iteritems
         elif hasattr(value, 'items'):
             return recursive(dict(value.items()), level=level + 1)
-        elif hasattr(value, '__iter__'):
+        elif hasattr(value, '__iter__') and not isinstance(value, io.IOBase):
             return list(map(recursive, value))
         elif convert_instances and hasattr(value, '__dict__'):
             # Likely an instance of something. Watch for cycles.
@@ -165,8 +171,14 @@ def to_primitive(value, convert_instances=False, convert_datetime=True,
     except TypeError:
         # Class objects are tricky since they may define something like
         # __iter__ defined but it isn't callable as list().
-        return six.text_type(value)
+        return fallback(value)
 
+    if orig_fallback is not None:
+        return orig_fallback(value)
+
+    # TODO(gcb) raise ValueError in version 3.0
+    warnings.warn("Cannot convert %r to primitive, will raise ValueError "
+                  "instead of warning in version 3.0" % (value,))
     return value
 
 
@@ -203,11 +215,7 @@ def dump_as_bytes(obj, default=to_primitive, encoding='utf-8', **kwargs):
 
     .. versionadded:: 1.10
     """
-    serialized = dumps(obj, default=default, **kwargs)
-    if isinstance(serialized, six.text_type):
-        # On Python 3, json.dumps() returns Unicode
-        serialized = serialized.encode(encoding)
-    return serialized
+    return dumps(obj, default=default, **kwargs).encode(encoding)
 
 
 def dump(obj, fp, *args, **kwargs):
